from .hdf5like_write_base import Hdf5LikeWrite
from gunpowder.coordinate import Coordinate
from gunpowder.ext import ZarrFile
from gunpowder.compat import ensure_str
import logging
import os
import traceback

logger = logging.getLogger(__name__)

class ZarrWrite(Hdf5LikeWrite):
    '''Assemble arrays of passing batches in one zarr container. This is useful
    to store chunks produced by :class:`Scan` on disk without keeping the
    larger array in memory. The ROIs of the passing arrays will be used to
    determine the position where to store the data in the dataset.

    Args:

        dataset_names (``dict``, :class:`ArrayKey` -> ``string``):

            A dictionary from array keys to names of the datasets to store them
            in.

        output_dir (``string``):

            The directory to save the zarr container. Will be created, if it does
            not exist.

        output_filename (``string``):

            The output filename of the container. Will be created, if it does
            not exist, otherwise data is overwritten in the existing container.

        compression_type (``string`` or ``int``):

            Compression strategy.  Legal values are ``gzip``, ``szip``,
            ``lzf``. If an integer between 1 and 10, this indicates ``gzip``
            compression level.

        dataset_dtypes (``dict``, :class:`ArrayKey` -> data type):

            A dictionary from array keys to datatype (eg. ``np.int8``). If
            given, arrays are stored using this type. The original arrays
            within the pipeline remain unchanged.
    '''

    def _get_voxel_size(self, dataset):

        if self.output_filename.endswith('.n5'):
            return Coordinate(dataset.attrs['resolution'][::-1])
        else:
            return Coordinate(dataset.attrs['resolution'])

    def _get_offset(self, dataset):

        if self.output_filename.endswith('.n5'):
            return Coordinate(dataset.attrs['offset'][::-1])
        else:
            return Coordinate(dataset.attrs['offset'])

    def _set_voxel_size(self, dataset, voxel_size):

        if self.output_filename.endswith('.n5'):
            dataset.attrs['resolution'] = voxel_size[::-1]
        else:
            dataset.attrs['resolution'] = voxel_size

    def _set_offset(self, dataset, offset):

        if self.output_filename.endswith('.n5'):
            dataset.attrs['offset'] = offset[::-1]
        else:
            dataset.attrs['offset'] = offset

    def _open_file(self, filename):
        return ZarrFile(ensure_str(filename), mode='a')



import copy
import logging
import numpy as np

from .batch_filter import BatchFilter
from gunpowder.array import ArrayKey
from gunpowder.coordinate import Coordinate
from gunpowder.points import PointsKey
from gunpowder.roi import Roi

logger = logging.getLogger(__name__)

class Pad(BatchFilter):
    '''Add a constant intensity padding around arrays of another batch 
    provider. This is useful if your requested batches can be larger than what 
    your source provides.

    Args:

        key (:class:`ArrayKey` or :class:`PointsKey`):

            The array or points set to pad.

        size (:class:`Coordinate` or ``None``):

            The padding to be added. If None, an infinite padding is added. If
            a coordinate, this amount will be added to the ROI in the positive
            and negative direction.

        value (scalar or ``None``):

            The value to report inside the padding. If not given, 0 is used.
            Only used for :class:`Array<Arrays>`.
    '''

    def __init__(self, key, size, value=None):

        self.key = key
        self.size = size
        self.value = value

    def setup(self):

        assert self.key in self.spec, (
            "Asked to pad %s, but is not provided upstream."%self.key)
        assert self.spec[self.key].roi is not None, (
            "Asked to pad %s, but upstream provider doesn't have a ROI for "
            "it."%self.key)

        spec = self.spec[self.key].copy()
        if self.size is not None:
            spec.roi = spec.roi.grow(self.size, self.size)
        else:
            spec.roi.set_shape(None)
        self.updates(self.key, spec)

    def prepare(self, request):

        upstream_spec = self.get_upstream_provider().spec

        logger.debug("request: %s"%request)
        logger.debug("upstream spec: %s"%upstream_spec)

        if self.key not in request:
            return

        roi = request[self.key].roi.copy()

        # change request to fit into upstream spec
        request[self.key].roi = roi.intersect(upstream_spec[self.key].roi)

        if request[self.key].roi.empty():

            logger.warning(
                "Requested %s ROI %s lies entirely outside of upstream "
                "ROI %s.", self.key, roi, upstream_spec[self.key].roi)

            # ensure a valid request by asking for empty ROI
            request[self.key].roi = Roi(
                    upstream_spec[self.key].roi.get_offset(),
                    (0,)*upstream_spec[self.key].roi.dims()
            )

        logger.debug("new request: %s"%request)

    def process(self, batch, request):

        if self.key not in request:
            return

        # restore requested batch size and ROI
        if isinstance(self.key, ArrayKey):

            array = batch.arrays[self.key]
            array.data = self.__expand(
                    array.data,
                    array.spec.roi/array.spec.voxel_size,
                    request[self.key].roi/array.spec.voxel_size,
                    self.value if self.value else 0
            )
            array.spec.roi = request[self.key].roi

        else:

            points = batch.points[self.key]
            points.spec.roi = request[self.key].roi

    def __expand(self, a, from_roi, to_roi, value):
        '''from_roi and to_roi should be in voxels.'''

        logger.debug(
            "expanding array of shape %s from %s to %s",
            str(a.shape), from_roi, to_roi)

        num_channels = len(a.shape) - from_roi.dims()
        channel_shapes = a.shape[:num_channels]

        b = np.zeros(channel_shapes + to_roi.get_shape(), dtype=a.dtype)
        if value != 0:
            b[:] = value

        shift = tuple(-x for x in to_roi.get_offset())
        logger.debug("shifting 'from' by " + str(shift))
        a_in_b = from_roi.shift(shift).to_slices()

        logger.debug("target shape is " + str(b.shape))
        logger.debug("target slice is " + str(a_in_b))

        b[(slice(None),)*num_channels + a_in_b] = a

        return b



from gunpowder.ext import caffe, NoSuchModule

try:
    class SolverParameters(caffe.SolverParameter):
        '''Wrapper of the caffe's ``SolverParameter`` class.

        Attributes:

            train_net (``string``):

                The network to train.

            base_lr (``float``):

                The initial learning rate.

            snapshot (``int``):

                Weight snapshot interval in iterations.

            snapshot_prefix (``string``)

                Prefix of snapshot files.

            type (``string``):

                Optimizer type, e.g., ``Adam``

                For ``Adam``, the following parameters can be set as attributes
                of this class: ``momentum`` (``float``), ``momentum2``
                (``float``), ``delta`` (``float``), ``weight_decay``
                (``float``), ``lr_policy`` (``string``), ``gamma`` (``float``),
                ``power`` (``float``).

            resume_from (``string`` or ``None``):

                Weight snapshot file to resume training from.

            train_state (training stages):

                Used to set the current trainig stage (if used during
                construction of the network).
        '''
        pass
except:
    class SolverParameters(NoSuchModule):
        '''Wrapper of the caffe's ``SolverParameter`` class.

        Attributes:

            train_net (``string``):

                The network to train.

            base_lr (``float``):

                The initial learning rate.

            snapshot (``int``):

                Weight snapshot interval in iterations.

            snapshot_prefix (``string``)

                Prefix of snapshot files.

            type (``string``):

                Optimizer type, e.g., ``Adam``

                For ``Adam``, the following parameters can be set as attributes
                of this class: ``momentum`` (``float``), ``momentum2``
                (``float``), ``delta`` (``float``), ``weight_decay``
                (``float``), ``lr_policy`` (``string``), ``gamma`` (``float``),
                ``power`` (``float``).

            resume_from (``string`` or ``None``):

                Weight snapshot file to resume training from.

            train_state (training stages):

                Used to set the current trainig stage (if used during
                construction of the network).
        '''
        def __init__(self):
            super(SolverParameters, self).__init__('caffe')



import logging
import multiprocessing
import numpy as np
from gunpowder.array import Array
from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.points import Points
from gunpowder.producer_pool import ProducerPool
from gunpowder.roi import Roi
from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)

class Scan(BatchFilter):
    '''Iteratively requests batches of size ``reference`` from upstream
    providers in a scanning fashion, until all requested ROIs are covered. If
    the batch request to this node is empty, it will scan the complete upstream
    ROIs (and return nothing). Otherwise, it scans only the requested ROIs and
    returns a batch assembled of the smaller requests. In either case, the
    upstream requests will be contained in the downstream requested ROI or
    upstream ROIs.

    See also :class:`Hdf5Write`.

    Args:

        reference (:class:`BatchRequest`):

            A reference :class:`BatchRequest`. This request will be shifted in
            a scanning fashion over the upstream ROIs of the requested arrays
            or points.

        num_workers (``int``, optional):

            If set to >1, upstream requests are made in parallel with that
            number of workers.

        cache_size (``int``, optional):

            If multiple workers are used, how many batches to hold at most.
    '''

    def __init__(self, reference, num_workers=1, cache_size=50):

        self.reference = reference.copy()
        self.num_workers = num_workers
        self.cache_size = cache_size
        self.workers = None
        if num_workers > 1:
            self.request_queue = multiprocessing.Queue(maxsize=0)
        self.batch = None

    def setup(self):

        if self.num_workers > 1:
            self.workers = ProducerPool(
                [self.__worker_get_chunk for _ in range(self.num_workers)],
                queue_size=self.cache_size)
            self.workers.start()

    def teardown(self):

        if self.num_workers > 1:
            self.workers.stop()

    def provide(self, request):

        empty_request = (len(request) == 0)
        if empty_request:
            scan_spec = self.spec
        else:
            scan_spec = request

        stride = self.__get_stride()
        shift_roi = self.__get_shift_roi(scan_spec)

        shifts = self.__enumerate_shifts(shift_roi, stride)
        num_chunks = len(shifts)

        logger.info("scanning over %d chunks", num_chunks)

        # the batch to return
        self.batch = Batch()

        if self.num_workers > 1:

            for shift in shifts:
                shifted_reference = self.__shift_request(self.reference, shift)
                self.request_queue.put(shifted_reference)

            for i in range(num_chunks):

                chunk = self.workers.get()

                if not empty_request:
                    self.__add_to_batch(request, chunk)

                logger.info("processed chunk %d/%d", i + 1, num_chunks)

        else:

            for i, shift in enumerate(shifts):

                shifted_reference = self.__shift_request(self.reference, shift)
                chunk = self.__get_chunk(shifted_reference)

                if not empty_request:
                    self.__add_to_batch(request, chunk)

                logger.info("processed chunk %d/%d", i + 1, num_chunks)

        batch = self.batch
        self.batch = None

        logger.debug("returning batch %s", batch)

        return batch

    def __get_stride(self):
        '''Get the maximal amount by which ``reference`` can be moved, such
        that it tiles the space.'''

        stride = None

        # get the least common multiple of all voxel sizes, we have to stride
        # at least that far
        lcm_voxel_size = self.spec.get_lcm_voxel_size(
            self.reference.array_specs.keys())

        # that's just the minimal size in each dimension
        for key, reference_spec in self.reference.items():

            shape = reference_spec.roi.get_shape()

            for d in range(len(lcm_voxel_size)):
                assert shape[d] >= lcm_voxel_size[d], ("Shape of reference "
                                                       "ROI %s for %s is "
                                                       "smaller than least "
                                                       "common multiple of "
                                                       "voxel size "
                                                       "%s"%(reference_spec.roi,
                                                             key,
                                                             lcm_voxel_size))

            if stride is None:
                stride = shape
            else:
                stride = Coordinate((
                    min(a, b)
                    for a, b in zip(stride, shape)))

        return stride

    def __get_shift_roi(self, spec):
        '''Get the minimal and maximal shift (as a ROI) to apply to
        ``self.reference``, such that it is still fully contained in ``spec``.
        '''

        total_shift_roi = None

        # get individual shift ROIs and intersect them
        for key, reference_spec in self.reference.items():

            logger.debug(
                "getting shift roi for %s with spec %s",
                key,
                reference_spec)

            if key not in spec:
                logger.debug("skipping, %s not in upstream spec", key)
                continue
            if spec[key].roi is None:
                logger.debug("skipping, %s has not ROI", key)
                continue

            logger.debug("upstream ROI is %s", spec[key].roi)

            for r, s in zip(
                    reference_spec.roi.get_shape(),
                    spec[key].roi.get_shape()):
                assert r <= s, (
                    "reference %s with ROI %s does not fit into provided "
                    "upstream %s"%(key, reference_spec.roi, spec[key].roi))

            # we have a reference ROI
            #
            #    [--------) [9]
            #    3        12
            #
            # and a spec ROI
            #
            #                 [---------------) [16]
            #                 16              32
            #
            # min and max shifts of reference are
            #
            #                 [--------) [9]
            #                 16       25
            #                        [--------) [9]
            #                        23       32
            #
            # therefore, all possible ways to shift the reference such that it
            # is contained in the spec are at least 16-3=13 and at most 23-3=20
            # (inclusive)
            #
            #              [-------) [8]
            #              13      21
            #
            # 1. the starting point is beginning of spec - beginning of reference
            # 2. the length is length of spec - length of reference + 1

            # 1. get the starting point of the shift ROI
            shift_begin = (
                spec[key].roi.get_begin() -
                reference_spec.roi.get_begin())

            # 2. get the shape of the shift ROI
            shift_shape = (
                spec[key].roi.get_shape() -
                reference_spec.roi.get_shape() +
                (1,)*reference_spec.roi.dims())

            # create a ROI...
            shift_roi = Roi(shift_begin, shift_shape)

            logger.debug("shift ROI for %s is %s", key, shift_roi)

            # ...and intersect it with previous shift ROIs
            if total_shift_roi is None:
                total_shift_roi = shift_roi
            else:
                total_shift_roi = total_shift_roi.intersect(shift_roi)
                if total_shift_roi.empty():
                    raise RuntimeError("There is no location where the ROIs "
                                       "the reference %s are contained in the "
                                       "request/upstream ROIs "
                                       "%s."%(self.reference, spec))

            logger.debug("intersected with total shift ROI this yields %s",
                    total_shift_roi)

        if total_shift_roi is None:
            raise RuntimeError("None of the upstream ROIs are bounded (all "
                               "ROIs are None). Scan needs at least one "
                               "bounded upstream ROI.")

        return total_shift_roi

    def __enumerate_shifts(self, shift_roi, stride):
        '''Produces a sequence of shift coordinates starting at the beginning
        of ``shift_roi``, progressing with ``stride``. The maximum shift
        coordinate in any dimension will be the last point inside the shift roi
        in this dimension.'''

        min_shift = shift_roi.get_offset()
        max_shift = Coordinate(m - 1 for m in shift_roi.get_end())

        shift = np.array(min_shift)
        shifts = []

        dims = len(min_shift)

        logger.debug(
            "enumerating possible shifts of %s in %s", stride, shift_roi)

        while True:

            logger.debug("adding %s", shift)
            shifts.append(Coordinate(shift))

            if (shift == max_shift).all():
                break

            # count up dimensions
            for d in range(dims):

                if shift[d] >= max_shift[d]:
                    if d == dims - 1:
                        break
                    shift[d] = min_shift[d]
                else:
                    shift[d] += stride[d]
                    # snap to last possible shift, don't overshoot
                    if shift[d] > max_shift[d]:
                        shift[d] = max_shift[d]
                    break

        return shifts

    def __shift_request(self, request, shift):

        shifted = request.copy()
        for _, spec in shifted.items():
            spec.roi = spec.roi.shift(shift)

        return shifted

    def __worker_get_chunk(self):

        request = self.request_queue.get()
        return self.__get_chunk(request)

    def __get_chunk(self, request):

        return self.get_upstream_provider().request_batch(request)

    def __add_to_batch(self, spec, chunk):

        if self.batch.get_total_roi() is None:
            self.batch = self.__setup_batch(spec, chunk)

        for (array_key, array) in chunk.arrays.items():
            if array_key not in spec:
                continue
            self.__fill(self.batch.arrays[array_key].data, array.data,
                        spec.array_specs[array_key].roi, array.spec.roi,
                        self.spec[array_key].voxel_size)

        for (points_key, points) in chunk.points.items():
            if points_key not in spec:
                continue
            self.__fill_points(self.batch.points[points_key].data, points.data,
                               spec.points_specs[points_key].roi, points.roi)

    def __setup_batch(self, batch_spec, chunk):
        '''Allocate a batch matching the sizes of ``batch_spec``, using
        ``chunk`` as template.'''

        batch = Batch()

        for (array_key, spec) in batch_spec.array_specs.items():
            roi = spec.roi
            voxel_size = self.spec[array_key].voxel_size

            # get the 'non-spatial' shape of the chunk-batch
            # and append the shape of the request to it
            array = chunk.arrays[array_key]
            shape = array.data.shape[:-roi.dims()]
            shape += (roi.get_shape() // voxel_size)

            spec = self.spec[array_key].copy()
            spec.roi = roi
            logger.info("allocating array of shape %s for %s", shape, array_key)
            batch.arrays[array_key] = Array(data=np.zeros(shape),
                                                spec=spec)

        for (points_key, spec) in batch_spec.points_specs.items():
            roi = spec.roi
            spec = self.spec[points_key].copy()
            spec.roi = roi
            batch.points[points_key] = Points(data={}, spec=spec)

        logger.debug("setup batch to fill %s", batch)

        return batch

    def __fill(self, a, b, roi_a, roi_b, voxel_size):
        logger.debug("filling " + str(roi_b) + " into " + str(roi_a))

        roi_a = roi_a // voxel_size
        roi_b = roi_b // voxel_size

        common_roi = roi_a.intersect(roi_b)
        if common_roi.empty():
            return

        common_in_a_roi = common_roi - roi_a.get_offset()
        common_in_b_roi = common_roi - roi_b.get_offset()

        slices_a = common_in_a_roi.get_bounding_box()
        slices_b = common_in_b_roi.get_bounding_box()

        if len(a.shape) > len(slices_a):
            slices_a = (slice(None),)*(len(a.shape) - len(slices_a)) + slices_a
            slices_b = (slice(None),)*(len(b.shape) - len(slices_b)) + slices_b

        a[slices_a] = b[slices_b]

    def __fill_points(self, a, b, roi_a, roi_b):
        logger.debug("filling points of " + str(roi_b) + " into points of" + str(roi_a))

        common_roi = roi_a.intersect(roi_b)
        if common_roi is None:
            return

        # find max point_id in a so far
        max_point_id = 0
        for point_id, point in a.items():
            if point_id > max_point_id:
                max_point_id = point_id

        for point_id, point in b.items():
            if roi_a.contains(Coordinate(point.location)):
                a[point_id + max_point_id] = point



import logging
import numpy as np

logger = logging.getLogger(__name__)

def get_net_input_specs(net):
    input_specs = {}
    for layer in net.layers:
        if layer.type == 'MemoryData':
            for i in range(0, layer.layer_param.top_size):
                blob_name = layer.layer_param.get_top(i)
                blob = net.blobs[blob_name]
                input_spec = InputSpec(blob_name, layer, blob, np.shape(blob.data))
                input_specs[input_spec.name] = input_spec
    return input_specs

def get_net_output_specs(net, blob_names):
    output_specs = {}
    for blob_name in blob_names:
        output_spec = OutputSpec(blob_name, net.blobs[blob_name], np.shape(net.blobs[blob_name].data))
        output_specs[output_spec.name] = output_spec
    return output_specs

class InputSpec(object):

    def __init__(self, name, memory_layer, blob, shape, phase=0):
        self.name = name
        self.memory_layer = memory_layer
        self.blob = blob
        self.shape = shape
        self.phase = phase # only added to the network

class OutputSpec(object):

    def __init__(self, name, blob, shape):
        self.name = name
        self.blob = blob
        self.shape = shape

# Wrapper around a networks set_input_arrays to prevent memory leaks of locked
# up arrays
class NetIoWrapper:

    def __init__(self, net, names_net_outputs):
        '''
        :param net:                network prototxt file
        :param names_net_outputs:  list, names of network outputs
        '''
        self.net = net
        self.input_specs = get_net_input_specs(net)
        self.output_specs = get_net_output_specs(net, names_net_outputs)
        self.inputs = {}

        for set_key in self.input_specs.keys():
            shape = self.input_specs[set_key].shape
            # Pre-allocate arrays that will persist with the network
            self.inputs[set_key] = np.zeros(tuple(shape), dtype=np.float32)

    def set_inputs(self, data):
        for set_key in self.input_specs.keys():
            try:
                np.copyto(self.inputs[set_key], np.ascontiguousarray(data[set_key]).astype(np.float32))
                self.net.set_layer_input_arrays(self.input_specs[set_key].memory_layer, self.inputs[set_key], None)
            except:
                logger.error("Could not set input '%s':"%set_key)
                raise

    def get_outputs(self):
        outputs = {}
        for set_key in self.output_specs.keys():
            outputs[set_key] = self.output_specs[set_key].blob.data
        return outputs

    def get_output_diffs(self):
        diffs = {}
        for set_key in self.output_specs.keys():
            diffs[set_key] = self.output_specs[set_key].blob.diff
        return diffs



import numpy as np
import logging
from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.nodes.batch_provider import BatchProvider
from gunpowder.points import Point, Points
from gunpowder.points_spec import PointsSpec
from gunpowder.profiling import Timing
from gunpowder.roi import Roi

logger = logging.getLogger(__name__)

class CsvPointsSource(BatchProvider):
    '''Read a set of points from a comma-separated-values text file. Each line
    in the file represents one point.

    Args:

        filename (``string``):

            The file to read from.

        points (:class:`PointsKey`):

            The key of the points set to create.

        points_spec (:class:`PointsSpec`, optional):

            An optional :class:`PointsSpec` to overwrite the points specs
            automatically determined from the CSV file. This is useful to set
            the :class:`Roi` manually.

        scale (scalar or array-like):

            An optional scaling to apply to the coordinates of the points read
            from the CSV file. This is useful if the points refer to voxel
            positions to convert them to world units.
    '''

    def __init__(self, filename, points, points_spec=None, scale=None):

        self.filename = filename
        self.points = points
        self.points_spec = points_spec
        self.scale = scale
        self.ndims = None
        self.data = None

    def setup(self):

        self._read_points()

        if self.points_spec is not None:

            self.provides(self.points, self.points_spec)
            return

        min_bb = Coordinate(np.floor(np.amin(self.data[:,:self.ndims], 0)))
        max_bb = Coordinate(np.ceil(np.amax(self.data[:,:self.ndims], 0)) + 1)

        roi = Roi(min_bb, max_bb - min_bb)

        self.provides(self.points, PointsSpec(roi=roi))

    def provide(self, request):

        timing = Timing(self)
        timing.start()

        min_bb = request[self.points].roi.get_begin()
        max_bb = request[self.points].roi.get_end()

        logger.debug(
            "CSV points source got request for %s",
            request[self.points].roi)

        point_filter = np.ones((self.data.shape[0],), dtype=np.bool)
        for d in range(self.ndims):
            point_filter = np.logical_and(point_filter, self.data[:,d] >= min_bb[d])
            point_filter = np.logical_and(point_filter, self.data[:,d] < max_bb[d])

        points_data = self._get_points(point_filter)
        points_spec = PointsSpec(roi=request[self.points].roi.copy())

        batch = Batch()
        batch.points[self.points] = Points(points_data, points_spec)

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def _get_points(self, point_filter):

        filtered = self.data[point_filter]
        ids = np.arange(len(self.data))[point_filter]

        return {
            i: Point(p)
            for i, p in zip(ids, filtered)
        }

    def _read_points(self):
        self.data, self.ndims = self._parse_csv()

    def _parse_csv(self, ndims=0):
        '''Read one point per line. If ``ndims`` is 0, all values in one line
        are considered as the location of the point. If positive, only the
        first ``ndims`` are used. If negative, all but the last ``-ndims`` are
        used.
        '''

        points = np.array(
            [
                [ float(t.strip(',')) for t in line.split() ]
                for line in open(self.filename, 'r')
            ], dtype=np.float32)

        if ndims == 0:
            ndims = points.shape[1]
        elif ndims < 0:
            ndims = points.shape[1] + ndims

        if self.scale is not None:
            points[:,:ndims] *= self.scale

        return points, ndims



import numpy as np

from .batch_filter import BatchFilter

class IntensityAugment(BatchFilter):
    '''Randomly scale and shift the values of an intensity array.

    Args:

        array (:class:`ArrayKey`):

            The intensity array to modify.

        scale_min (``float``):
        scale_max (``float``):
        shift_min (``float``):
        shift_max (``float``):

            The min and max of the uniformly randomly drawn scaling and
            shifting values for the intensity augmentation. Intensities are
            changed as::

                a = a.mean() + (a-a.mean())*scale + shift

        z_section_wise (``bool``):

            Perform the augmentation z-section wise. Requires 3D arrays and
            assumes that z is the first dimension.
    '''

    def __init__(self, array, scale_min, scale_max, shift_min, shift_max, z_section_wise=False):
        self.array = array
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.shift_min = shift_min
        self.shift_max = shift_max
        self.z_section_wise = z_section_wise

    def process(self, batch, request):

        raw = batch.arrays[self.array]

        assert not self.z_section_wise or raw.spec.roi.dims() == 3, "If you specify 'z_section_wise', I expect 3D data."
        assert raw.data.dtype == np.float32 or raw.data.dtype == np.float64, "Intensity augmentation requires float types for the raw array (not " + str(raw.data.dtype) + "). Consider using Normalize before."
        assert raw.data.min() >= 0 and raw.data.max() <= 1, "Intensity augmentation expects raw values in [0,1]. Consider using Normalize before."

        if self.z_section_wise:
            for z in range((raw.spec.roi/self.spec[self.array].voxel_size).get_shape()[0]):
                raw.data[z] = self.__augment(
                        raw.data[z],
                        np.random.uniform(low=self.scale_min, high=self.scale_max),
                        np.random.uniform(low=self.shift_min, high=self.shift_max))
        else:
            raw.data = self.__augment(
                    raw.data,
                    np.random.uniform(low=self.scale_min, high=self.scale_max),
                    np.random.uniform(low=self.shift_min, high=self.shift_max))

        # clip values, we might have pushed them out of [0,1]
        raw.data[raw.data>1] = 1
        raw.data[raw.data<0] = 0

    def __augment(self, a, scale, shift):

        return a.mean() + (a-a.mean())*scale + shift
